"""High performance copula generators
"""
from typing import List, Optional, Union, cast

import numpy as np
import numpy.random
import numpy.linalg
import dask.array as da
from dask.array.core import normalize_chunks

from . import duck
from .sobol import sobol
from .typing import Chunks2D, NormalizedChunks2D


def gaussian_copula(cov: Union[List[List[float]], np.ndarray],
                    samples: int, seed: int = 0,
                    chunks: Chunks2D = None,
                    rng: str = 'Mersenne Twister'
                    ) -> Union[np.ndarray, da.Array]:
    """Gaussian Copula scenario generator.

    Simplified algorithm::

        >>> l = numpy.linalg.cholesky(cov)
        >>> y = numpy.random.standard_normal(size=(samples, cov.shape[0]))
        >>> p = numpy.dot(l, y.T).T

    :param numpy.ndarray cov:
        covariance matrix, a.k.a. correlation matrix. It must be a
        Hermitian, positive-definite matrix in any square array-like format.
        The width of cov determines the number of dimensions of the output.

    :param int samples:
        Number of random samples to generate

        .. note::
           When using Sobol, to obtain a uniform distribution one must use
           :math:`2^{n} - 1` samples (for any n > 0).

    :param chunks:
        Chunk size for the return array, which has shape (samples, dimensions).
        It can be anything accepted by dask (a positive integer, a tuple of two
        ints, or a tuple of two tuples of ints) for the output shape.

        Set to None to return a numpy array.

        .. warning::
           When using the Mersenne Twister random generator, the chunk size
           changes the random sequence. To guarantee repeatability, it must be
           fixed together with the seed. chunks=None also produces different
           results from using dask.

    :param int seed:
        Random seed.

        With ``rng='Sobol'``, this is the initial dimension; when generating
        multiple copulas with different seeds, one should never use seeds that
        are less than ``cov.shape[0]`` apart from each other.

        The maximum seed when using sobol is::

            pysamples.sobol.max_dimensions() - cov.shape[0] - 1

    :param str rng:
        Either ``Mersenne Twister`` or ``Sobol``

    :returns:
        array of shape (samples, dimensions), with all series
        being normal (0, 1) distributions.
    :rtype:
        If chunks is not None, :class:`dask.array.Array`; else
        :class:`numpy.ndarray`
    """
    assert samples > 0
    cov = np.asarray(cov)
    assert cov.ndim == 2
    assert cov.shape[0] == cov.shape[1]

    L = numpy.linalg.cholesky(cov)  # type: Union[np.ndarray, da.Array]
    if chunks:
        chunks = cast(NormalizedChunks2D,
                      normalize_chunks(chunks, shape=(samples, cov.shape[0])))
        L = da.from_array(L, chunks=(chunks[1], chunks[1]))

    rng = rng.lower()
    if rng == 'mersenne twister':
        rnd_state = duck.RandomState(seed)
        # When pulling samples from the Mersenne Twister generator, we have
        # the samples on the rows. This guarantees that if we draw more
        # samples, the original samples won't change.
        y = rnd_state.standard_normal(size=(samples, cov.shape[0]),
                                      chunks=chunks)
    elif rng == 'sobol':
        # Generate uniform (0, 1) distributions
        samples = sobol(size=(samples, cov.shape[0]),
                        d0=seed, chunks=chunks)
        # Convert to normal (0, 1)
        y = duck.norm_ppf(samples)
    else:
        raise ValueError("Unknown rng: %s" % rng)

    return duck.dot(L, y.T).T


def t_copula(cov: Union[List[List[float]], np.ndarray],
             df: Union[int, List[int], np.ndarray],
             samples: int, seed: int = 0,
             chunks: Chunks2D = None,
             rng: str = 'Mersenne Twister'
             ) -> Union[np.ndarray, da.Array]:
    """Student T Copula / IT Copula scenario generator.

    Simplified algorithm::

        >>> l = numpy.linalg.cholesky(cov)
        >>> y = numpy.random.standard_normal(size=(samples, cov.shape[0]))
        >>> r = numpy.random.uniform(size=(samples, 1))
        >>> s = scipy.stats.chi2.ppf(r, df=df)
        >>> z = numpy.sqrt(df / s) * numpy.dot(l, y.T).T
        >>> u = scipy.stats.t.cdf(z, df=df)
        >>> p = scipy.stats.norm.ppf(u)

    :param df:
        Number of degrees of freedom. Can be either a scalar int for
        Student T Copula, or a one-dimensional array-like with one point per
        dimension for IT Copula.

    :param int seed:
        Random seed.

        With ``rng='Sobol'``, this is the initial dimension; when generating
        multiple copulas with different seeds, one should never use seeds that
        are less than ``cov.shape[0] + 1`` apart from each other.

        The maximum seed when using sobol is::

            pysamples.sobol.max_dimensions() - cov.shape[0] - 2

    All other parameters and the return value are the same as in
    :func:`gaussian_copula`.
    """
    assert samples > 0
    cov = np.asarray(cov)
    assert cov.ndim == 2
    assert cov.shape[0] == cov.shape[1]
    dimensions = cov.shape[0]

    L = numpy.linalg.cholesky(cov)
    if chunks is not None:
        chunks = cast(NormalizedChunks2D,
                      normalize_chunks(chunks, shape=(samples, dimensions)))
        L = da.from_array(L, chunks=(chunks[1], chunks[1]))

    # Pre-process df into a 1D dask array
    df = np.asarray(df)
    if (df <= 0).any():
        raise ValueError("df must always be greater than zero")
    if df.shape not in ((), (dimensions, )):
        raise ValueError("df must be either a scalar or a 1D vector with as "
                         "many points as the width of the correlation matrix")
    if df.ndim == 1 and chunks is not None:
        df = da.from_array(df, chunks=(chunks[1], ))

    # Define chunks for the S chi-square matrix
    chunks_r = None  # type: Optional[NormalizedChunks2D]
    if chunks is not None:
        chunks_r = (chunks[0], (1, ))

    rng = rng.lower()
    if rng == 'mersenne twister':
        # Use two separate random states for the normal and the chi2
        # distributions. This is NOT the same as just extracting two series
        # from the same RandomState, as we must guarantee that, if you extract
        # a different number of samples from the generator, the initial
        # samples must remain the same.
        # For the same reason, we have the samples on the rows.
        rnd_state_y = duck.RandomState(seed)
        # Don't just do seed + 1 as that would have unwanted repercussions
        # when one tries to extract different series from different seeds.
        seed_r = (seed + 190823761298456) % 2**32
        rnd_state_r = duck.RandomState(seed_r)

        y = rnd_state_y.standard_normal(size=(samples, dimensions),
                                        chunks=chunks)
        r = rnd_state_r.uniform(size=(samples, 1), chunks=chunks_r)

    elif rng == 'sobol':
        seed_r = seed + dimensions

        y = sobol(size=(samples, dimensions), d0=seed, chunks=chunks)
        y = duck.norm_ppf(y)
        r = sobol(size=(samples, 1), d0=seed_r, chunks=chunks_r)

    else:
        raise ValueError("Unknown rng: %s" % rng)

    s = duck.chi2_ppf(r, df)
    z = duck.sqrt(df / s) * duck.dot(L, y.T).T
    # Convert t distribution to normal (0, 1)
    u = duck.t_cdf(z, df)
    p = duck.norm_ppf(u)
    return p



import pytest
import numpy as np
import scipy.stats
from numpy.testing import assert_allclose
from pyscenarios.copula import gaussian_copula, t_copula
from pyscenarios.stats import tail_dependence


cov = [[1.0, 0.9, 0.7],
       [0.9, 1.0, 0.4],
       [0.7, 0.4, 1.0]]


def test_gaussian_mersenne_np():
    actual = gaussian_copula(cov, samples=4, seed=123, chunks=None,
                             rng='Mersenne Twister')
    expect = [[-1.08563060, -0.54233474, -1.15002017],
              [-1.50629471, -1.60787125,  0.04561073],  # noqa
              [-2.42667924, -2.37097000, -0.86315499],
              [-0.86674040, -1.07598597, -0.29407627]]
    assert_allclose(expect, actual, 1e-6, 0)
    assert isinstance(actual, np.ndarray)


def test_gaussian_mersenne_da():
    actual = gaussian_copula(cov, samples=4, seed=123, chunks=2,
                             rng='Mersenne Twister')
    expect = [[ 1.61739599,  1.34668479,  0.85965138],  # noqa
              [ 0.00936110,  0.77307164, -0.65639660],  # noqa
              [-0.88414686, -1.07582330, -0.47076357],
              [-1.63883139, -1.21794817, -1.47348213]]
    assert_allclose(expect, actual, 1e-6, 0)
    assert actual.chunks == ((2, 2), (2, 1))


@pytest.mark.parametrize('chunks,expect_chunks', [
    (None, None),
    (2, ((2, 2), (2, 1))),
    (((3, 1), (2, 1)), ((3, 1), (2, 1))),
])
def test_gaussian_sobol(chunks, expect_chunks):
    actual = gaussian_copula(cov, samples=4, seed=123, chunks=chunks,
                             rng='Sobol')
    expect = [[ 0.        ,  0.        ,  0.        ],  # noqa
              [-0.67448975, -0.31303751, -1.15262386],
              [ 0.67448975,  0.31303751,  1.15262386],  # noqa
              [ 0.31863936,  0.78820110, -0.53727912]]  # noqa
    assert_allclose(expect, actual, 1e-6, 0)
    if chunks:
        assert actual.chunks == expect_chunks
    else:
        assert isinstance(actual, np.ndarray)


def test_student_t_mersenne_np():
    actual = t_copula(cov, df=3, samples=4, seed=123, chunks=None,
                      rng='Mersenne Twister')
    expect = [[-1.48810072, -0.86441534, -1.54668921],
              [-1.29120595, -1.35445918,  0.04614266],  # noqa
              [-1.69841381, -1.67427423, -0.76713426],
              [-0.52427748, -0.64222262, -0.18205599]]
    assert_allclose(expect, actual, 1e-6, 0)
    assert isinstance(actual, np.ndarray)


def test_student_t_mersenne_da():
    actual = t_copula(cov, df=3, samples=4, seed=123, chunks=2,
                      rng='Mersenne Twister')
    expect = [[ 0.9251499 ,  0.78838765,  0.52075254],  # noqa
              [ 0.00733126,  0.58581369, -0.50181461],  # noqa
              [-0.46342698, -0.55862319, -0.25037537],
              [-0.92150594, -0.70869994, -0.84036731]]
    assert_allclose(expect, actual, 1e-6, 0)
    assert actual.chunks == ((2, 2), (2, 1))


@pytest.mark.parametrize('chunks,expect_chunks', [
    (None, None),
    (2, ((2, 2), (2, 1))),
    (((3, 1), (2, 1)), ((3, 1), (2, 1))),
])
def test_student_t_sobol(chunks, expect_chunks):
    actual = t_copula(cov, df=3, samples=4, seed=123, chunks=chunks,
                      rng='Sobol')
    expect = [[ 0.        ,  0.        ,  0.        ],  # noqa
              [-0.90292647, -0.44513114, -1.38033019],
              [ 0.51756147,  0.24504617,  0.84650386],  # noqa
              [ 0.59093028,  1.28328308, -0.94456215]]  # noqa
    assert_allclose(expect, actual, 1e-6, 0)
    if chunks:
        assert actual.chunks == expect_chunks
    else:
        assert isinstance(actual, np.ndarray)


def test_it_mersenne_np():
    actual = t_copula(cov, df=[3, 4, 5], samples=4, seed=123, chunks=None,
                      rng='Mersenne Twister')
    expect = [[-1.48810072, -0.80579141, -1.48680149],
              [-1.29120595, -1.40226642,  0.04565698],  # noqa
              [-1.69841381, -1.77537961, -0.79048734],
              [-0.52427748, -0.68508399, -0.20095568]]
    assert_allclose(expect, actual, 1e-6, 0)
    assert isinstance(actual, np.ndarray)


def test_it_mersenne_da():
    actual = t_copula(cov, df=[3, 4, 5], samples=4, seed=123,
                      chunks=2, rng='Mersenne Twister')
    expect = [[ 0.92514990,  0.84367196,  0.57859631],  # noqa
              [ 0.00733126,  0.60760124, -0.53208793],  # noqa
              [-0.46342698, -0.60402557, -0.28417274],
              [-0.92150594, -0.75838980, -0.94757843]]
    assert_allclose(expect, actual, 1e-6, 0)
    assert actual.chunks == ((2, 2), (2, 1))


@pytest.mark.parametrize('chunks,expect_chunks', [
    (None, None),
    (2, ((2, 2), (2, 1))),
    (((3, 1), (2, 1)), ((3, 1), (2, 1))),
])
def test_it_sobol(chunks, expect_chunks):
    actual = t_copula(cov, df=[3, 4, 5], samples=4, seed=123,
                      chunks=chunks, rng='Sobol')
    expect = [[ 0.        ,  0.        ,  0.        ],  # noqa
              [-0.90292647, -0.41928686, -1.35361744],
              [ 0.51756147,  0.25248047,  0.91032037],  # noqa
              [ 0.59093028,  1.20943444, -0.81940488]]  # noqa
    assert_allclose(expect, actual, 1e-6, 0)
    if chunks:
        assert actual.chunks == expect_chunks
    else:
        assert isinstance(actual, np.ndarray)


all_copulas = pytest.mark.parametrize(
    'func,kwargs', [
        (gaussian_copula, {'rng': 'Mersenne Twister'}),
        (gaussian_copula, {'rng': 'Mersenne Twister', 'chunks': (4096, 2)}),
        (gaussian_copula, {'rng': 'Sobol'}),
        (t_copula, {'df': 8, 'rng': 'Mersenne Twister'}),
        (t_copula, {'df': 8, 'rng': 'Mersenne Twister', 'chunks': (4096, 2)}),
        (t_copula, {'df': 8, 'rng': 'Sobol'}),
        (t_copula, {'df': [8, 9, 10], 'rng': 'Mersenne Twister'}),
        (t_copula, {'df': [8, 9, 10], 'rng': 'Mersenne Twister',
                    'chunks': (4096, 2)}),
        (t_copula, {'df': [8, 9, 10], 'rng': 'Sobol'}),
    ])


@all_copulas
def test_normal_01(func, kwargs):
    """All copulas produce normal (0, 1) distributions for all series
    """
    s = func(cov, samples=65535, **kwargs)
    s = scipy.stats.norm.cdf(s)
    hist, bin_edges = np.histogram(s)

    assert_allclose(hist / 65535, [.3] * 10, rtol=0, atol=1e-2)
    assert_allclose(bin_edges, [0, .1, .2, .3, .4, .5, .6, .7, .8, .9, 1],
                    rtol=0, atol=1e-2)


@all_copulas
def test_extra_samples(func, kwargs):
    """Drawing additional samples from a copula does not change the
    realised sequence of the previous samples
    """
    s1 = func(cov, samples=5000, **kwargs)
    s2 = func(cov, samples=9000, **kwargs)
    assert_allclose(s1, s2[:5000], atol=0, rtol=1e-12)


@all_copulas
def test_cov_roundtrip(func, kwargs):
    s = func(cov, samples=65535, **kwargs)
    cov2 = np.corrcoef(s.T)
    assert_allclose(cov, cov2, rtol=0, atol=0.05)


@pytest.mark.parametrize('df,expect_td', [
    (3, [.33, .33, .33]),
    (9, [.20, .20, .20]),
    (999, [.13, .13, .13]),
    ([3, 3, 999, 999], [.33, .08, .13])
])
@pytest.mark.parametrize('rng', ['Mersenne Twister', 'Sobol'])
@pytest.mark.parametrize('chunks', [None, (65536, 1)])
def test_tail_dependence(df, expect_td, rng, chunks):
    cov2 = [[1.0, 0.5, 0.5, 0.5],
            [0.5, 1.0, 0.5, 0.5],
            [0.5, 0.5, 1.0, 0.5],
            [0.5, 0.5, 0.5, 1.0]]
    s = t_copula(cov2, df=df, samples=262143, rng=rng, chunks=chunks)
    s = scipy.stats.norm.cdf(s)
    actual_td = [
        tail_dependence(s[:, 0], s[:, 1], .99),
        tail_dependence(s[:, 1], s[:, 2], .99),
        tail_dependence(s[:, 2], s[:, 3], .99),
    ]
    assert_allclose(expect_td, actual_td, atol=0.02, rtol=0)



"""Duck-typed functions that call numpy or dask depending on the inputs
"""
from functools import wraps
from typing import Any, Callable, Optional, Tuple, Union

import dask.array as da
import numpy as np
import scipy.stats
from .typing import Chunks2D


def array(x: Any) -> Union[np.ndarray, da.Array]:
    """Convert x to numpy array, unless it's a da.array
    """
    if isinstance(x, (np.ndarray, da.Array)):
        return x
    return np.array(x)


def _map_blocks(func: Callable[..., np.ndarray]
                ) -> Callable[..., Union[np.ndarray, da.Array]]:
    """Wrap an arbitrary function that takes one or more arrays in input.
    If any is a Dask Array, invoke :func:`dask.array.map_blocks`, otherwise
    apply the function directly.
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        if any(isinstance(arg, da.Array) for arg in args):
            return da.map_blocks(func, *args, **kwargs, dtype=args[0].dtype)
        return func(*args, **kwargs)
    return wrapper


def _map_blocks_df(func: Callable[[Any, Any], np.ndarray]
                   ) -> Callable[[Any, Any], Union[np.ndarray, da.Array]]:
    """Specialized variant for functions with degrees of freedom - adds
    auto-chunking in case of mismatched arguments
    """
    @wraps(func)
    def wrapper(x, df):
        x = array(x)
        df = array(df)
        if isinstance(x, da.Array) or isinstance(df, da.Array):
            # map_blocks auto-broadcasting broken since dask 1.1
            # https://github.com/dask/dask/issues/4739
            x, df = da.broadcast_arrays(x, df)
            return da.map_blocks(func, x, df, dtype=float)
        return func(x, df)
    return wrapper


def _toplevel(func_name: str) -> Callable[..., Union[np.ndarray, da.Array]]:
    """If any of the args is a Dask Array, invoke da.func_name; else invoke
    np.func_name
    """
    def wrapper(*args, **kwargs):
        if any(isinstance(arg, da.Array) for arg in args):
            func = getattr(da, func_name)
        else:
            func = getattr(np, func_name)
        return func(*args, **kwargs)
    return wrapper


norm_cdf = _map_blocks(scipy.stats.norm.cdf)
norm_ppf = _map_blocks(scipy.stats.norm.ppf)
chi2_cdf = _map_blocks_df(scipy.stats.chi2.cdf)
chi2_ppf = _map_blocks_df(scipy.stats.chi2.ppf)
t_cdf = _map_blocks_df(scipy.stats.t.cdf)
t_ppf = _map_blocks_df(scipy.stats.t.ppf)
dot = _toplevel('dot')
sqrt = _toplevel('sqrt')
where = _toplevel('where')


class RandomState:
    """Wrapper around :class:`numpy.random.RandomState` and
    :class:`dask.array.random.RandomState`.

    For each method, if chunks=None invoke the numpy version, otherwise invoke
    the dask version.
    """
    def __init__(self, seed: Optional[int] = None):
        self._dask_state = da.random.RandomState(seed)

    @property
    def _numpy_state(self) -> np.random.RandomState:
        return self._dask_state._numpy_state

    def seed(self, seed: Optional[int] = None) -> None:
        self._dask_state.seed(seed)

    def _apply(self, func_name: str, size: Optional[Tuple[int, int]] = None,
               chunks: Chunks2D = None):
        if chunks is not None:
            func = getattr(self._dask_state, func_name)
            return func(size=size, chunks=chunks)
        else:
            func = getattr(self._numpy_state, func_name)
            return func(size=size)

    def uniform(self, size: Optional[Tuple[int, int]] = None,
                chunks: Chunks2D = None):
        return self._apply('uniform', size=size, chunks=chunks)

    def standard_normal(self, size: Optional[Tuple[int, int]] = None,
                        chunks: Chunks2D = None):
        return self._apply('standard_normal', size=size, chunks=chunks)



import dask.array as da
import numpy as np
import pytest
from numpy.testing import assert_array_equal
from pyscenarios.stats import tail_dependence


@pytest.mark.parametrize('chunk', [False, True])
def test_tail_dependence(chunk):
    x = [.1, .3, .4, .5, .6, .8, .9]
    y = [.4, .1, .2, .5, .9, .9, .1]
    q = [[.05, .15, .35, .50, .75],
         [.35, .75, .50, .70, .05]]

    if chunk:
        x = da.from_array(np.array(x), chunks=5)
        y = da.from_array(np.array(y), chunks=5)
        q = da.from_array(np.array(q), chunks=(1, 2))

    d = tail_dependence(x, y, q)
    assert_array_equal(d, [[np.nan, 0., 0.5, 0.75, 0.5],
                           [.5, .5, .75, .5, np.nan]])

    if chunk:
        assert d.chunks == q.chunks
    else:
        assert isinstance(d, np.ndarray)

    # Scalar q
    d = tail_dependence(x, y, .5)
    assert_array_equal(d, .75)

    if chunk:
        assert d.chunks == ()
    else:
        assert isinstance(d, float)



"""Sobol sequence generator

This is a reimplementation of a C++ algorithm by
`Stephen Joe and Frances Y. Kuo <http://web.maths.unsw.edu.au/~fkuo/sobol/>`_.
Directions are based on :file:`new-joe-kuo-6.21201` from the URL above.
"""
import lzma
import pkg_resources
from typing import Iterator, Tuple, Union, cast

import numpy as np
import dask.array as da
from dask.array.core import normalize_chunks
from numba import jit

from .typing import Chunks2D, NormalizedChunks2D

__all__ = ('sobol', 'max_dimensions')

DIRECTIONS = 'new-joe-kuo-6.21201.txt.xz'


v_cache = None


def load_v() -> np.ndarray:
    """Load V from the original author's file. This function is executed
    automatically the first time you call the :func:`sobol` function.
    When using a dask backend, the V array is only loaded when
    actually needed by the kernel; this results in smaller pickle files.
    When using dask distributed, V is loaded locally on the workers instead of
    being transferred over the network.
    """
    global v_cache
    if v_cache is None:
        with pkg_resources.resource_stream('pyscenarios', DIRECTIONS) as fh:
            with lzma.open(fh, 'rt') as zfh:
                directions = _load_directions(zfh)
        v_cache = _calc_v_kernel(directions)
    return v_cache


def _load_directions(fh: Iterator[str]) -> np.ndarray:
    """Load input file containing direction numbers.
    The file must one of those available on the website of the
    original author, or formatted like one.

    :returns:
        contents of the input file as a 2D numpy array.
        Every row contains the information for the matching dimension.
        Column 0 contains the a values, while columns 1+ contain the m values.
        The m values are padded on the right with zeros.
    """
    rows = [row.split() for row in fh]

    # Add padding at end of rows
    # Drop first 2 columns
    # Replace header with element for d=1
    rowlen = len(rows[-1])
    for row in rows:
        row[:] = row[2:] + ['0'] * (rowlen - len(row))
    rows[0] = ['0'] + ['1'] * (rowlen - 3)
    return np.array(rows, dtype='uint32')


@jit('uint32[:,:](uint32[:,:])', nopython=True, nogil=True, cache=True)
def _calc_v_kernel(directions: np.ndarray) -> np.ndarray:
    """Numba kernel for :func:`calc_v`
    """
    # Initialise temp array of direction numbers
    v = np.empty((directions.shape[0], 32), dtype=np.uint32)

    for j in range(directions.shape[0]):
        s = 0
        # Compute direction numbers
        for s in range(directions.shape[1] - 1):
            if directions[j, s + 1] == 0:
                break
            v[j, s] = directions[j, s + 1] * 2**(31 - s)
        else:
            # need a C-style for loop
            # for(s=0; s<m.size; s++)
            # where at the end of the loop s == m.size
            s += 1

        for t in range(s, 32):
            v[j, t] = v[j, t - s] ^ (v[j, t - s] // 2**s)
            for k in range(1, s):
                v[j, t] ^= (
                    ((directions[j, 0] // 2**(s - 1 - k)) & 1)
                    * v[j, t - k])

    return v


def _sobol_kernel(samples: int, dimensions: int, s0: int, d0: int
                  ) -> np.ndarray:
    """Numba kernel for :func:`sobol`

    :returns:
        points 2D array.

        points[i, j] = the jth component of the ith point
        with i indexed from 0 to N-1 and j indexed from 0 to D-1
    """
    output = np.empty((samples, dimensions), order='F')
    _sobol_kernel_jit(samples, dimensions, s0, d0, load_v(), output)
    return output


@jit('void(uint32, uint32, uint32, uint32, uint32[:, :], float64[:, :])',
     nopython=True, nogil=True, cache=True)
def _sobol_kernel_jit(samples: int, dimensions: int, s0: int, d0: int,
                      V: np.ndarray, output: np.ndarray) -> None:
    """Jit-compiled core of :func:`_sobol_kernel

    When running in dask and there are multiple chunks on the
    samples, calculate and then discard the first s0 samples.
    This is inefficient but preferable to transferring a state
    vector across the graph, which would introduce cross-chunks dependencies.
    """
    for j in range(dimensions):
        state = 0
        for i in range(s0 + samples):
            # c = index from the right of the first zero bit of i
            c = 0
            mask = 1
            while i & mask:
                mask *= 2
                c += 1

            state ^= V[j + d0, c]
            if i >= s0:
                output[i - s0, j] = np.double(state) / np.double(2**32)


def sobol(size: Union[int, Tuple[int, int]], d0: int = 0,
          chunks: Chunks2D = None) -> Union[np.ndarray, da.Array]:
    """Sobol points generator based on Gray code order

    :param size:
        number of samples (cannot be greater than :math:`2^{32}`) to extract
        from a single dimension, or tuple (samples, dimensions).
        To guarantee uniform distribution, the number of samples should
        always be :math:`2^{n} - 1`.
    :param int d0:
        first dimension. This can be used as a functional equivalent of a
        a random seed. dimensions + d0 can't be greater than
        :func:`max_dimensions()` - 1.
    :param chunks:
        If None, return a numpy array.

        If set, return a dask array with the given chunk size.
        It can be anything accepted by dask (a positive integer, a
        tuple of two ints, or a tuple of two tuples of ints) for the output
        shape (see result below). e.g. either ``(16384, 50)`` or
        ``((16384, 16383),  (50, 50, 50))`` could be used together with
        ``size=(32767, 150)``.

        .. note::
           The algorithm is not efficient if there are multiple chunks on axis
           0. However, if you do need them, it is typically better to require
           them here than re-chunking afterwards, particularly if (most of) the
           subsequent algorithm is embarassingly parallel.
    :returns:
        If size is an int, a 1-dimensional array of samples.
        If size is a tuple, a 2-dimensional array POINTS, where
        ``POINTS[i, j]`` is the ith sample of the jth dimension.
        Each dimension is a uniform (0, 1) distribution.
    :rtype:
        If chunks is not None, :class:`dask.array.Array`; else
        :class:`numpy.ndarray`
    """
    if isinstance(size, int):
        samples = size
        dimensions = 1
    else:
        samples, dimensions = size

    if not 0 < samples < 2**32:
        raise ValueError("samples must be between 1 and 2^32")
    if not 0 < dimensions + d0 <= max_dimensions():
        raise ValueError("(dimensions + d0) must be between 1 and %d" %
                         max_dimensions())

    if chunks is None:
        res = _sobol_kernel(samples, dimensions, 0, d0)
        if isinstance(size, int):
            res = res[:, 0]
        return res

    # dask-specific code
    chunks = cast(NormalizedChunks2D,
                  normalize_chunks(chunks, shape=(samples, dimensions)))
    name = 'sobol-%d-%d-%d' % (samples, dimensions, d0)
    dsk = {}

    offset_i = 0
    for i, size_i in enumerate(chunks[0]):
        offset_j = 0
        for j, size_j in enumerate(chunks[1]):
            dsk[name, i, j] = (_sobol_kernel, size_i, size_j,
                               offset_i, d0 + offset_j)
            offset_j += size_j
        offset_i += size_i

    res = da.Array(dsk, name=name, dtype=float, chunks=chunks)
    if isinstance(size, int):
        res = res[:, 0]
    return res


def max_dimensions() -> int:
    """Return number of dimensions available. When invoking :func:`sobol`,
    ``size[1] + d0`` must be smaller than this.
    """
    return load_v().shape[0]



try:
    from .version import version as __version__  # noqa: F401
except ImportError:  # pragma: no cover
    raise ImportError('pyscenarios not properly installed. If you are running '
                      'from the source directory, please instead create a new '
                      'virtual environment (using conda or virtualenv) and '
                      'then install it in-place by running pip install -e .')



import pytest
import dask.array as da
import numpy as np
import scipy.stats
from numpy.testing import assert_array_equal, assert_allclose, assert_equal
from pyscenarios import duck


def test_array():
    for x in (1, [1, 2], np.array(1), np.array([1, 2])):
        y = duck.array(x)
        assert isinstance(y, np.ndarray)
        assert_equal(x, y)

    x = da.arange(3, chunks=2)
    y = duck.array(x)
    assert x is y


@pytest.mark.parametrize('chunk', [False, True])
@pytest.mark.parametrize('func,wrapped', [
    (duck.norm_cdf, scipy.stats.norm.cdf),
    (duck.norm_ppf, scipy.stats.norm.ppf),
    (duck.sqrt, np.sqrt),
])
def test_map_blocks(func, wrapped, chunk):
    x = np.random.rand(10)
    y = wrapped(x)

    if chunk:
        dx = da.from_array(x, chunks=5)
    else:
        dx = x
    dy = func(dx)

    assert_array_equal(y, dy)
    if chunk:
        assert dy.chunks == ((5, 5), )
    else:
        assert isinstance(dy, np.ndarray)


@pytest.mark.parametrize('df,chunk_df', [
    (3, False),
    ([1, 2, 3], False),
    (np.array([1, 2, 3]), False),
    (np.array([1, 2, 3]), True)
])
@pytest.mark.parametrize('x,chunk_x', [
    (np.random.rand(4).reshape(4, 1), False),
    (np.random.rand(4).reshape(4, 1), True),
    # dask fails to broadcast a np.ndarray with shape[i] != 1
    # against a da.Array with len(chunks[i]) != 1
    pytest.param(np.random.rand(30).reshape(10, 3), False,
                 marks=pytest.mark.xfail),
    pytest.param(np.random.rand(30).reshape(10, 3), True,
                 marks=pytest.mark.xfail),
    pytest.param([[1, 2, 3], [4, 5, 6]], False,
                 marks=pytest.mark.xfail),
])
@pytest.mark.parametrize('func,wrapped', [
    (duck.chi2_cdf, scipy.stats.chi2.cdf),
    (duck.chi2_ppf, scipy.stats.chi2.ppf),
    (duck.t_cdf, scipy.stats.t.cdf),
    (duck.t_ppf, scipy.stats.t.ppf),
])
def test_map_blocks_df(func, wrapped, x, chunk_x, df, chunk_df):
    y = wrapped(x, df)

    if chunk_x:
        dx = da.from_array(x, chunks=2)
    else:
        dx = x
    if chunk_df:
        ddf = da.from_array(df, chunks=2)
    else:
        ddf = df

    dy = func(dx, ddf)

    assert_array_equal(y, dy)
    if chunk_x or chunk_df:
        assert isinstance(dy, da.Array)
    else:
        assert isinstance(dy, np.ndarray)


@pytest.mark.parametrize('chunk', [False, True])
def test_dot(chunk):
    x = np.random.rand(16).reshape(4, 4)
    y = np.random.rand(24).reshape(4, 6)
    z = np.dot(x, y)

    if chunk:
        dx = da.from_array(x, chunks=2)
        dy = da.from_array(y, chunks=2)
    else:
        dx = x
        dy = y
    dz = duck.dot(dx, dy)
    assert_allclose(z, dz, 1e-12, 0)
    if chunk:
        assert dz.chunks == ((2, 2), (2, 2, 2))
    else:
        assert isinstance(dz, np.ndarray)


@pytest.mark.parametrize('chunk', [False, True])
def test_where(chunk):
    x = np.array([1, 2, 3])
    y = np.array([0, 5, 2])
    z = duck.where(x > y, x, y)

    if chunk:
        dx = da.from_array(x, chunks=2)
        dy = da.from_array(y, chunks=2)
    else:
        dx = x
        dy = y
    dz = duck.where(dx > dy, dx, dy)
    assert_array_equal(z, dz)
    if chunk:
        assert dz.chunks == ((2, 1),)
    else:
        assert isinstance(dz, np.ndarray)


@pytest.mark.parametrize('chunks', [None, 2, ((2, 1), (2, 1))])
def test_randomstate_uniform(chunks):
    state = duck.RandomState(123)
    ref_np = np.random.RandomState(123)
    ref_da = da.random.RandomState(123)

    seq = state.uniform(size=(3, 3), chunks=chunks)
    if chunks:
        assert_array_equal(
            seq, ref_da.uniform(size=(3, 3), chunks=chunks))
        assert seq.chunks == ((2, 1), (2, 1))
    else:
        assert_array_equal(seq, ref_np.uniform(size=(3, 3)))
        assert isinstance(seq, np.ndarray)


@pytest.mark.parametrize('chunks', [None, 2, ((2, 1), (2, 1))])
def test_randomstate_standard_normal(chunks):
    state = duck.RandomState(123)
    ref_np = np.random.RandomState(123)
    ref_da = da.random.RandomState(123)

    seq = state.standard_normal(size=(3, 3), chunks=chunks)
    if chunks:
        assert_array_equal(
            seq, ref_da.standard_normal(size=(3, 3), chunks=chunks))
        assert seq.chunks == ((2, 1), (2, 1))
    else:
        assert_array_equal(seq, ref_np.standard_normal(size=(3, 3)))
        assert isinstance(seq, np.ndarray)



"""Statistical functions
"""
from typing import Any, Union
import dask.array as da
import numpy as np
from . import duck


def tail_dependence(x: Any, y: Any, q: Any) -> Union[np.ndarray, da.Array]:
    r"""Calculate `tail dependence
    <https://en.wikipedia.org/wiki/Tail_dependence>`_
    between vectors x and y.

    :param x:
        1D array-like or dask array containing samples from a
        uniform (0, 1) distribution.
    :param y:
        other array to compare against
    :param q:
        quantile(s) (0 < q < 1).
        Either a scalar or a ND array-like or dask array.
    :returns:
        array of the same shape and type as q, containing:

        .. math::

            \cases{
                P(y < q | x < q) | q < 0.5 \cr
                P(y \geq q | x \geq q) | q \geq 0.5
            }
    """
    x = duck.array(x)
    y = duck.array(y)
    q = duck.array(q)

    assert x.size == y.size
    assert x.ndim == y.ndim == 1

    q = q[..., np.newaxis]
    x_lt_q = x < q
    x_ge_q = x >= q
    xcount = duck.where(q < .5, x_lt_q, x_ge_q)
    ltail = x_lt_q & (y < q)
    htail = x_ge_q & (y >= q)
    tail = duck.where(q < .5, ltail, htail)

    return tail.sum(axis=-1) / xcount.sum(axis=-1)



from typing import Tuple, Union

Chunks2D = Union[
    None, int, Tuple[int, int],
    Tuple[Tuple[int, ...], Tuple[int, ...]]]
"""Chunk size for a return array, which has shape (samples, dimensions).
It can be anything accepted by dask (a positive integer, a tuple of two
ints, or a tuple of two tuples of ints) for the output shape or None to
request a numpy array.
"""
NormalizedChunks2D = Tuple[Tuple[int, ...], Tuple[int, ...]]



#!/usr/bin/env python
import os
import re
import warnings
from setuptools import find_packages, setup


MAJOR = 0
MINOR = 2
MICRO = 0
ISRELEASED = True
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)
QUALIFIER = ''
DISTNAME = 'pyscenarios'
LICENSE = 'Apache'
AUTHOR = 'Guido Imperiale'
AUTHOR_EMAIL = 'guido.imperiale@gmail.com'
URL = 'https://github.com/crusaderky/pyscenarios'
CLASSIFIERS = [
    'Development Status :: 3 - Alpha',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
    'Intended Audience :: Science/Research',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Topic :: Scientific/Engineering',
]

INSTALL_REQUIRES = [
    'dask >= 0.17.3',
    'numba >= 0.34',
    'numpy >= 1.13',
    'pandas >= 0.20',
    'scipy >= 1.0',
]
TESTS_REQUIRE = ['pytest >= 3.6']

DESCRIPTION = "Python Monte Carlo Scenario Generator"
LONG_DESCRIPTION = DESCRIPTION

# Code to extract and write the version copied from pandas.
# Used under the terms of pandas's license, see licenses/PANDAS_LICENSE.
FULLVERSION = VERSION
write_version = True

if not ISRELEASED:
    import subprocess
    FULLVERSION += '.dev'

    pipe = None
    for cmd in ['git', 'git.cmd']:
        try:
            pipe = subprocess.Popen(
                [cmd, "describe", "--always", "--match", "v[0-9]*"],
                stdout=subprocess.PIPE)
            (so, serr) = pipe.communicate()
            if pipe.returncode == 0:
                break
        except BaseException:
            pass

    if pipe is None or pipe.returncode != 0:
        # no git, or not in git dir
        if os.path.exists('pyscenarios/version.py'):
            warnings.warn(
                "WARNING: Couldn't get git revision,"
                " using existing pyscenarios/version.py")
            write_version = False
        else:
            warnings.warn(
                "WARNING: Couldn't get git revision,"
                " using generic version string")
    else:
        # have git, in git dir, but may have used a shallow clone (travis does
        # this)
        rev = so.strip()
        rev = rev.decode('ascii')

        if not rev.startswith('v') and re.match("[a-zA-Z0-9]{7,9}", rev):
            # partial clone, manually construct version string
            # this is the format before we started using git-describe
            # to get an ordering on dev version strings.
            rev = "v%s+dev.%s" % (VERSION, rev)

        # Strip leading v from tags format "vx.y.z" to get th version string
        FULLVERSION = rev.lstrip('v')

        # make sure we respect PEP 440
        FULLVERSION = FULLVERSION.replace("-", "+dev", 1).replace("-", ".")

else:
    FULLVERSION += QUALIFIER


def write_version_py(filename=None):
    cnt = """\
version = '%s'
short_version = '%s'
"""
    if not filename:
        filename = os.path.join(
            os.path.dirname(__file__), 'pyscenarios', 'version.py')

    a = open(filename, 'w')
    try:
        a.write(cnt % (FULLVERSION, VERSION))
    finally:
        a.close()


if write_version:
    write_version_py()


setup(name=DISTNAME,
      version=FULLVERSION,
      license=LICENSE,
      author=AUTHOR,
      author_email=AUTHOR_EMAIL,
      classifiers=CLASSIFIERS,
      description=DESCRIPTION,
      long_description=LONG_DESCRIPTION,
      install_requires=INSTALL_REQUIRES,
      tests_require=TESTS_REQUIRE,
      python_requires='>=3.5.0',
      url=URL,
      package_data={'pyscenarios': ['*.txt.xz']},
      packages=find_packages())
